import { useEffect, useState, useRef, MouseEvent } from 'react';
import { Canvas as WaveformCanvas } from './Canvas';
import {
  CanvasLogicalWidth,
  CanvasLogicalHeight,
  Selection,
} from '../Waveform';
import { mouseEventToCanvasX } from './Helpers';

type Props = {
  peaks: number[][] | null;
  numFrames: number;
  style: React.CSSProperties;
  onSelectionChange: (selection: Selection) => void;
};

enum Mode {
  Normal,
  Selecting,
}

export const Waveform: React.FC<Props> = (props: Props) => {
  const hudCanvasRef = useRef<HTMLCanvasElement>(null);
  const [mode, setMode] = useState(Mode.Normal);

  const defaultSelection: Selection = { x1: 0, x2: 0 };
  const [selection, setSelection] = useState(defaultSelection);
  // TODO: is it needed?
  const [newStartSelection, setNewStartSelection] = useState(0);

  // effects

  useEffect(() => {
    (async function () {
      const canvas = hudCanvasRef.current;
      if (canvas == null) {
        console.error('no hud canvas ref available');
        return;
      }

      const ctx = canvas.getContext('2d');
      if (ctx == null) {
        console.error('no hud 2d context available');
        return;
      }

      ctx.clearRect(0, 0, canvas.width, canvas.height);

      if (selection.x1 >= selection.x2) {
        return;
      }

      ctx.beginPath();
      ctx.strokeStyle = 'red';
      ctx.lineWidth = 2;
      ctx.fillStyle = 'rgba(255, 255, 255, 0.3)';
      ctx.rect(selection.x1, 2, selection.x2 - selection.x1, canvas.height - 8);
      ctx.fill();
      ctx.stroke();
    })();
  });

  // handlers

  const handleMouseDown = (evt: MouseEvent<HTMLCanvasElement>) => {
    console.log('mousedown');
    if (mode != Mode.Normal) {
      return;
    }

    setMode(Mode.Selecting);
    setNewStartSelection(mouseEventToCanvasX(evt));
  };

  const handleMouseMove = (evt: MouseEvent<HTMLCanvasElement>) => {
    console.log('mousemove');
    if (mode != Mode.Selecting) {
      return;
    }

    const selection: Selection = {
      x1: newStartSelection,
      x2: mouseEventToCanvasX(evt),
    };
    setSelection(selection);
  };

  const handleMouseUp = () => {
    console.log('mouseup');
    if (mode != Mode.Selecting) {
      return;
    }
    setMode(Mode.Normal);
    props.onSelectionChange(selection);
  };

  // render component

  const canvasStyles = {
    width: '100%',
    height: '100%',
    margin: '0 auto',
    display: 'block',
  } as React.CSSProperties;

  const hudCanvasStyles = {
    width: '100%',
    height: '100%',
    position: 'absolute',
    top: 0,
    left: 0,
    right: 0,
    bottom: 0,
    zIndex: 1,
  } as React.CSSProperties;

  return (
    <>
      <div style={props.style}>
        <WaveformCanvas
          peaks={props.peaks}
          fillStyle="grey"
          strokeStyle="black"
          style={canvasStyles}
        ></WaveformCanvas>
        <canvas
          ref={hudCanvasRef}
          style={hudCanvasStyles}
          width={CanvasLogicalWidth}
          height={CanvasLogicalHeight}
          onMouseDown={handleMouseDown}
          onMouseMove={handleMouseMove}
          onMouseUp={handleMouseUp}
        ></canvas>
      </div>
    </>
  );
};
